"""
    by_locator        : tuple --> (<selenium By object>, <selector string>)
    x_offset          : int   --> integer value of x offset in pixels
    y_offset          : int   --> integer value of y offset in pixels
    x_destination     : int   -->  integer value of x location on page
    y_desitination    : int   --> integer value of y location on page
    by_locator_source : tuple --> (<selenium By object>, <selector string>)
    by_locator_target : tuple --> (<selenium By object>, <selector string>)
    clear_first       : bool  --> toggle for clearing input field before writing text to it
    press_enter       : bool  --> toggle for sending the ENTER key to an input field after writing to it
"""

from selenium.webdriver import ActionChains
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.keys import Keys
from po_utils.common_actions.waits import wait_for_page_to_load, wait_until_displayed, wait_until_not_displayed


@wait_until_displayed
@wait_for_page_to_load
def click_element(self, by_locator:tuple, x_offset:int=0, y_offset:int=0) -> None:
    # I hate clicking
    element = WebDriverWait(self._driver, self._driver_wait_time).until(EC.visibility_of_element_located(by_locator))
    scroll_height = self._driver.execute_script('return document.body.scrollHeight')
    window_size = self._driver.get_window_size()['height']
    if element.location['y'] > (scroll_height - .5 * window_size):
        self._driver.execute_script('window.scrollTo(0, document.body.scrollHeight)')
    elif element.location['y'] < (.5 * window_size):
        self._driver.execute_script('window.scrollTo(0, 0)')
    else:
        self._driver.execute_script(f"window.scrollTo({element.location['x']}, {element.location['y'] - .5 * window_size});")
    if x_offset == 0 and y_offset == 0:
        try:
            WebDriverWait(self._driver, self._driver_wait_time).until(EC.element_to_be_clickable(by_locator)).click()
        except:
            WebDriverWait(self._driver, self._driver_wait_time).until(EC.element_to_be_clickable(by_locator)).click()
    else:
        ActionChains(self._driver).move_to_element_with_offset(WebDriverWait(self._driver, self._driver_wait_time).until(EC.visibility_of_element_located(by_locator)), x_offset, y_offset).click().perform()


@wait_until_displayed
@wait_for_page_to_load
def click_and_drag_element_by_offset(self, by_locator:tuple, x_destination:int, y_desitination:int) -> None:
    element = WebDriverWait(self._driver, self._driver_wait_time).until(EC.visibility_of_element_located(by_locator))
    ActionChains(self._driver).drag_and_drop_by_offset(element, x_destination, y_desitination).perform()


@wait_until_displayed
@wait_for_page_to_load
def click_and_drag_element(self, by_locator_source:tuple, by_locator_target:tuple) -> None:
    source = WebDriverWait(self._driver, self._driver_wait_time).until(EC.visibility_of_element_located(by_locator_source))
    target = WebDriverWait(self._driver, self._driver_wait_time).until(EC.visibility_of_element_located(by_locator_target))
    ActionChains(self._driver).drag_and_drop(source, target).perform()


@wait_until_displayed
@wait_for_page_to_load
def send_text_to_element(self, by_locator:tuple, text:str, clear_first:bool=True, press_enter:bool=False) -> None:
    if clear_first:
        self._driver.find_element(*by_locator).clear()
    self._driver.find_element(*by_locator).send_keys(text)
    if press_enter:
        self._driver.find_element(*by_locator).send_keys(Keys.ENTER)


@wait_until_displayed
@wait_for_page_to_load
def hover_over_element(self, by_locator:tuple) -> None:
    element = self._driver.find_element(*by_locator)
    ActionChains(self._driver).move_to_element(element).perform()



"""
    by_locator        : tuple --> (<selenium By object>, <selector string>)
    attribute         : str   --> 'attribute of an html element'
    text              : str   --> 'text of the element'
    is_case_sensitive : bool  --> boolean to toggle case sensitivity
"""

from po_utils.common_actions.waits import wait_for_page_to_load, wait_until_displayed


@wait_until_displayed
@wait_for_page_to_load
def has_attribute(self, by_locator: tuple, attribute: str) -> bool:
    return self._driver.find_element(*by_locator).get_attribute(attribute) != None


@wait_until_displayed
@wait_for_page_to_load
def has_text(self, by_locator: tuple, text: str, is_case_sensitive:bool=False) -> bool:
    element_text = self._driver.find_element(*by_locator).text

    if is_case_sensitive:
        return element_text == text
    return element_text.lower() == text.lower()


@wait_for_page_to_load
def is_visible(self, by_locator: tuple) -> bool:
    try:
        el = self._driver.find_element(*by_locator)
        return el.is_displayed()
    except Exception:
        return False



from functools import wraps
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC


def wait_for_page_to_load(function):
    @wraps(function)
    def wrapper(self, *args, **kwargs):
        try:
            WebDriverWait(self.driver, self.__driver_wait_time).until(lambda _: self.driver.execute_script('return document.readyState') == 'complete')
        except Exception as e:
            pass
        function_call = function(self, *args, **kwargs)
        try:
            WebDriverWait(self.driver, self.__driver_wait_time).until(lambda _: self.driver.execute_script('return document.readyState') == 'complete')
        except Exception as e:
            pass
        return function_call
    return wrapper


def wait_until_displayed(function):
    @wraps(function)
    def wrapper(self, *args, **kwargs):
        try:
            WebDriverWait(self.driver, self.__driver_wait_time).until(EC.visibility_of_element_located(args[0]))
        except:
            pass
        return function(self, *args, **kwargs)
    return wrapper


def wait_until_not_displayed(function):
    @wraps(function)
    def wrapper(self, *args, **kwargs):
        try:
            WebDriverWait(self.driver, self.__driver_wait_time).until(EC.invisibility_of_element_located(args[0]))
        except:
            pass
        return function(self, *args, **kwargs)
    return wrapper




from po_utils.common_actions import element_interactions, gets, booleans
from re import compile, search


# helpers -----------------------------------------------
def __get_locator_by_re(locators, pattern):
    pattern = compile(pattern) if type(pattern) is str else pattern
    return next((loc for loc in locators.__dict__ if search(loc, pattern)), None)
# -------------------------------------------------------


class Component:
    def __init__(self, driver, wait_time=5):
        self._driver = driver
        self._driver_wait_time = wait_time
        self._driver.implicitly_wait(self._driver_wait_time)
        self.locators = Locator()
    
    # common element interactions -----------------------------------------------
    click_element =                     element_interactions.click_element
    click_and_drag_element_by_offset =  element_interactions.click_and_drag_element_by_offset
    click_and_drag_element =            element_interactions.click_and_drag_element
    hover_over_element =                element_interactions.hover_over_element
    send_text_to_element =              element_interactions.send_text_to_element

    # common gets -----------------------------------------------
    get_elements =          gets.get_elements
    get_element =           gets.get_element
    get_element_attribute = gets.get_element_attribute
    get_element_text =      gets.get_element_text
    
    # common booleans -----------------------------------------------
    has_attribute = booleans.has_attribute
    has_text =      booleans.has_text
    is_visible =    booleans.is_visible
    

class Locator:
    def __add__(self, other):
        combination = {}
        combination.update(self.__dict__)
        combination.update(other.__dict__)
        new = self.__class__()
        new.__dict__.update(combination)
        return new

    def __iadd__(self, other):
        combination = {}
        combination.update(self.__dict__)
        combination.update(other.__dict__)
        new = self.__class__()
        new.__dict__.update(combination)
        return new


class Modal:
    def __init__(self, driver, wait_time=5):
        self._driver = driver
        self._driver_wait_time = wait_time
        self._driver.implicitly_wait(self._driver_wait_time)
        self.locators = Locator()
    
    # modal specific -----------------------------------------------
    def close(self):
        locator = __get_locator_by_re(self.locators, r'(close.*button|exit.*button)')
        if locator:
            element_interactions.click_element(locator)
        else:
            raise AttributeError('No locator for the close/exit button')
    exit = close

    def confirm(self):
        locator = __get_locator_by_re(self.locators, r'(confirm.*button|okay.*button)')
        if locator:
            element_interactions.click_element(locator)
        else:
            raise AttributeError('No locator for the confirm/okay button')
    okay = confirm

    def decline(self):
        locator = __get_locator_by_re(self.locators, r'(decline.*button|cancel.*button)')
        if locator:
            element_interactions.click_element(locator)
        else:
            raise AttributeError('No locator for the cancel/decline button')
    cancel = decline

    # common element interactions -----------------------------------------------
    click_element =                     element_interactions.click_element
    click_and_drag_element_by_offset =  element_interactions.click_and_drag_element_by_offset
    click_and_drag_element =            element_interactions.click_and_drag_element
    hover_over_element =                element_interactions.hover_over_element
    send_text_to_element =              element_interactions.send_text_to_element

    # common gets -----------------------------------------------
    get_elements =                      gets.get_elements
    get_element =                       gets.get_element
    get_element_attribute =             gets.get_element_attribute
    get_element_text =                  gets.get_element_text
    
    # common booleans -----------------------------------------------
    has_attribute =                     booleans.has_attribute
    has_text =                          booleans.has_text
    is_visible =                        booleans.is_visible


class Page:
    def __init__(self, driver, wait_time=5):
        self._driver = driver
        self._driver_wait_time = wait_time
        self._driver.implicitly_wait(self._driver_wait_time)

    # browser-specific controls -----------------------------------------------
    def get_current_url(self):
        return self._driver.current_url

    def get_current_title(self):
        return self._driver.title

    def go_to_url(self, url):
        self._driver.get(url)

    def refresh(self) -> None:
        self._driver.refresh()

    def go_back(self) -> None:
        self._driver.back()

    def go_forward(self) -> None:
        self._driver.forward()

    def scroll_to_position_on_page(self, x_position, y_position):
        self._driver.execute_script(f'window.scrollTo({x_position}, {y_position})')
    
    def switch_to_tab(self, tab_position):
        self._driver.switch_to.window(self._driver.window_handles[tab_position])

    def switch_to_frame(self, frame_reference=None):
        self._driver.switch_to.frame(frame_reference) if frame_reference else self._driver.switch_to.default_content()

    def quit(self):
        self._driver.quit()

    # common element interactions -----------------------------------------------
    click_element =                     element_interactions.click_element
    click_and_drag_element_by_offset =  element_interactions.click_and_drag_element_by_offset
    click_and_drag_element =            element_interactions.click_and_drag_element
    hover_over_element =                element_interactions.hover_over_element
    send_text_to_element =              element_interactions.send_text_to_element

    # common gets -----------------------------------------------
    get_elements =                      gets.get_elements
    get_element =                       gets.get_element
    get_element_attribute =             gets.get_element_attribute
    get_element_text =                  gets.get_element_text

    # common booleans -----------------------------------------------
    has_attribute =                     booleans.has_attribute
    has_text =                          booleans.has_text
    is_visible =                        booleans.is_visible
    


"""
    by_locator : tuple --> (<selenium By object>, <selector string>)
    attribute  : str   --> 'attribute of an html element'
"""

from po_utils.common_actions.waits import wait_for_page_to_load, wait_until_displayed, wait_until_not_displayed


@wait_until_displayed
@wait_for_page_to_load
def get_elements(self, by_locator: tuple) -> object:
    return self._driver.find_elements(*by_locator)

@wait_until_displayed
@wait_for_page_to_load
def get_element(self, by_locator: tuple) -> object:
    res = get_elements(self, by_locator)[0]
    return res if res else []

@wait_until_displayed
@wait_for_page_to_load
def get_element_attribute(self, by_locator: tuple, attribute: str) -> object:
    return self._driver.find_element(*by_locator).get_attribute(attribute)

@wait_until_displayed
@wait_for_page_to_load
def get_element_text(self, by_locator: tuple) -> object:
    return self._driver.find_element(*by_locator).text
